from os.path import join
import sys
import os
import random
import shutil

try:
    import configparser
except ImportError:
    # Python 2.x fallback
    import ConfigParser as configparser
from time import time

# COMPSs/PyCOMPSs imports
from pycompss.api.task import task
from pycompss.api.binary import binary
from pycompss.api.constraint import constraint
from pycompss.api.mpi import mpi
from pycompss.api.api import compss_wait_on
from pycompss.api.api import compss_barrier
from pycompss.api.api import TaskGroup
from pycompss.api.api import compss_barrier_group
from pycompss.api.parameter import *
from pycompss.api.on_failure import on_failure
from pycompss.api.exceptions import COMPSsException

# project imports
from esm_ensemble_init import *
from hecuba import StorageObj
from hecuba_lib.esm_dynamic_analysis_results import *


@on_failure(management='IGNORE')
@mpi(binary="${FESOM_EXE}", runner="mpirun", processes="${FESOM_CORES}", working_dir="{{working_dir_exe}}",
     fail_by_exit_value=True)
@task(log_file={Type: FILE_OUT, StdIOStream: STDOUT}, working_dir_exe={Type: INOUT, Prefix: "#"})
def esm_simulation(log_file, working_dir_exe):
    pass


###def esm_member_checkpoint(exp_id, sdate, checkpoint_log):
###   #retrieve from Hecuba the last status of the ensemble members produced by the analysis (running in parallel)
###   #TODO: remove mocked value after getting these from hecuba
###   ensemble_status = esm_dynamic_analysis_results.get_by_alias(exp_id + "_esm_dynamic_analysis")
###   return ensemble_status.results[sdate]

@task(returns=bool)
def esm_member_checkpoint(exp_id, sdate):
    # retrieve from Hecuba the last status of the ensemble members produced by the analysis (running in parallel)
    print("Checking status member - " + sdate)
    ensemble_status = esm_dynamic_analysis_results.get_by_alias(exp_id + "_esm_dynamic_analysis")
    return bool(ensemble_status.results[sdate])


def esm_member(exp_id, sdate, n_sims, top_working_dir):
    try:
        for sim in range(1, n_sims + 1):
            working_dir_exe = top_working_dir + "/" + sdate
            log = working_dir_exe + "/" + "fesom2_" + str(exp_id) + "_" + str(sdate) + "_" + str(sim) + ".out"
            print("################## launching simulation " + sdate + "." + str(
                sim) + "in " + working_dir_exe + "######################")
            compss_wait_on(esm_simulation(log, working_dir_exe))
            # check the state of the member, for the first one there is nothing to check
            to_continue = compss_wait_on(esm_member_checkpoint(exp_id, sdate))
            # if it should be aborted we throug an exception
            if not to_continue:
                raise COMPSsException("Member diverged - Aborting member " + str(sdate))

    except COMPSsException:
        # React to the exception (maybe calling other tasks or with other parameters)
        # TODO: add logic to remove all data from working dirs/output folders
        esm_member_disposal(exp_id, sdate, top_working_dir)
        pass


###def esm_member_disposal(exp_id, sdate, top_working_dir):

###   # TODO: remove hecuba data aswell of the concerned aborted member
###   # path
###   path = os.path.join("/home/bsc32/bsc32044/results/output_core2/", sdate)
###   # removing directory
###   shutil.rmtree(path)

@on_failure(management='IGNORE')
@task(returns=bool)
def esm_member_disposal(exp_id, sdate, top_working_dir):
    # TODO: remove hecuba data aswell of the concerned aborted member
    # path
    path = os.path.join("/home/bsc32/bsc32044/results/output_core2/", sdate)
    # removing directory
    shutil.rmtree(path)
    return True


def esm_ensemble(exp_id):
    # 1 - Creation of group
    # try:
    with TaskGroup(str(exp_id), False):
        # 3 - Launch the ensemble members
        top_working_dir = exp_settings['common']['top_working_dir']
        sdates_list = (exp_settings['common']['ensemble_start_dates']).split()
        n_sims = int(exp_settings['common']['chunks'])
        print("We have " + str(n_sims) + " chunks ")

        for sdate in sdates_list:
            # working_dir_exe = top_working_dir + "/" + sdate
            # log = working_dir_exe + "/" + "fesom2_" + str(exp_id) + "_" + str(sdate) + ".out"
            # print("################################################### launching simulation in " + working_dir_exe + "###################################################")
            # esm_simulation(log, working_dir_exe)
            # todo: add parameter to load the number of simulation chunks to run (based on frecuency)
            esm_member(exp_id, sdate, n_sims, top_working_dir)

    compss_barrier_group(str(exp_id))
    # create flag file to stop the parallel process
    run_finished_flag = str(exp_id) + ".finished"
    open(top_working_dir + "/" + run_finished_flag, 'a').close()


# except COMPSsException:
# React to the exception (maybe calling other tasks or with other parameters)
# TODO: add logic to remove all data from working dirs/output folders
#   print("################# aborted sim ####################")
#   pass

# dummy method to test data exchange with Hecuba
@task(returns=bool)
def esm_dynamic_analysis(exp_id):
    try:

        print("######################## performing dynamic analysis for experiment " + exp_id + "###################")
        # create a dummy object
        # TODO: here is the launching point of the analysis, it will be a PyCOMPSs task
        ds = esm_dynamic_analysis_results()
        ds.results["1948"] = True
        ds.results["1958"] = True
        ds.results["1968"] = True
        ds.make_persistent(str(exp_id) + "_esm_dynamic_analysis")

    except COMPSsException:
        pass


if __name__ == "__main__":
    print("Running FESOM2 - Pycompss")
    # 0 - create a ramdon nr for the experiment id
    exp_id = str(sys.argv[1])

    # create dummy data in Hecuba
    esm_dynamic_analysis(exp_id)

    # prepare folder structure/configuration
    exp_settings = compss_wait_on(esm_ensemble_init(exp_id, True))
    # exp_settings =  compss_wait_on(exp_settings)
    print("##################################### Initialization completed ####################################")
    # run the experiment
    esm_ensemble(exp_id)

